/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/09/14 16:49:15
 */

import {
  css,
  html,
  Component,
  bind,
  range,
  nextTick,
  outsideClick,
  clearOutsideClick
} from 'wkit'
import ICONS from './svg.js'
import '../form/input.js'
import '../form/button.js'

const ACTTION = {
  bold: 'bold',
  italic: 'italic',
  under: 'underline',
  delete: 'strikeThrough',
  left: 'justifyLeft',
  center: 'justifyCenter',
  right: 'justifyRight',
  image: 'insertImage',
  font: 'fontSize',
  color: 'foreColor',
  link: 'createLink',
  ordered: 'insertOrderedList',
  unordered: 'insertUnorderedList'
}

const TOOLBAR = [
  'font',
  'color',
  'bold',
  'italic',
  'under',
  'delete',
  'ordered',
  'unordered',
  'table',
  'left',
  'center',
  'right',
  'link',
  'image',
  'fullscreen',
  'copy'
]

const COLORS = [
  '#f3f5fb',
  '#dae1e9',
  '#62778d',
  '#58d68d',
  '#3fc2a7',
  '#52a3de',
  '#ac61ce',
  '#ffb618',
  '#e67e22',
  '#ff5061',
  '#ff0000',
  '#000000'
]

// 获取一维数组转二维的行
function getY(i) {
  return (i / 9) >> 0
}
//获取一维数组转二维的列
function getX(i) {
  return i % 9
}

function getIndex(x, y) {
  return x + y * 9
}

class Editor extends Component {
  static watches = ['value']

  static props = {
    readonly: {
      type: Boolean,
      observer(v) {
        this.#updateStat()
      }
    },
    disabled: {
      type: Boolean,
      observer(v) {
        this.#updateStat()
      }
    }
  }

  static styles = [
    css`
      :host {
        display: flex;
        min-width: 200px;
        max-height: 720px;
        border-radius: 3px;
        transition: box-shadow 0.15s linear;
        background: var(--wc-editor-background, #fff);
      }
      table {
        width: 100%;
        border-spacing: 0;
        border-collapse: collapse;

        tr {
          background: #fff;
        }
        thead tr {
          background: var(--color-plain-1);
        }
        th,
        td {
          padding: 6px 13px;
          border: 1px solid var(--color-plain-2);
          vertical-align: middle;
        }
        th {
          font-weight: bold;
        }
        tr:nth-child(2n) {
          background: #fcfdff;
        }
      }

      ul,
      ol {
        margin-left: 1em;
      }
      a {
        color: var(--color-teal-1);
      }
      .noselect {
        -webkit-touch-callout: none;
        -webkit-user-select: none;
        -moz-user-select: none;
        user-select: none;
      }
    `,

    css`
      .editor {
        position: relative;
        flex: 1;
        display: flex;
        flex-direction: column;
        width: 100%;
        border: 1px solid var(--wc-editor-border-color, var(--color-grey-2));
        border-radius: inherit;
        font-size: 14px;
      }
      .toolbar {
        display: flex;
        height: 34px;
        padding: 5px;
        line-height: 24px;
        border-bottom: 1px solid var(--color-grey-1);

        span {
          position: relative;
          overflow: hidden;
          display: flex;
          justify-content: center;
          align-items: center;
          width: 24px;
          height: 24px;
          margin: 0 3px;
          border-radius: 3px;

          input {
            position: absolute;
            width: 100%;
            height: 100%;
            opacity: 0;
          }
          .icon {
            overflow: hidden;
            width: 70%;
            height: 70%;
            fill: currentColor;
            color: #62778d;
          }

          &:hover {
            background: #f7f8fb;
          }
        }
      }

      .wrapper {
        overflow: hidden;
        overflow-y: auto;
        flex: 1;
        padding: 6px 12px;

        &::-webkit-scrollbar {
          width: 6px;
        }

        &::-webkit-scrollbar-thumb {
          visibility: hidden;
          border-radius: 3px;
        }

        &:hover::-webkit-scrollbar-thumb {
          visibility: visible;
          background: rgba(0, 0, 0, 0.3);
        }
      }

      .typearea {
        width: 100%;
        height: 100%;
        min-height: 54px;
        line-height: 1.5;
        outline: none;
        text-wrap: wrap;
        word-break: break-all;

        img {
          max-width: 100%;
        }
      }
    `,

    css`
      :host([readonly]) {
        .editor {
          cursor: default;
          opacity: 0.8;
        }
      }
      :host([disabled]) {
        .editor {
          cursor: not-allowed;
          opacity: 0.6;
        }
      }
      :host([readonly]),
      :host([disabled]) {
        .toolbar {
          span:hover {
            background: none;
          }
        }
      }

      :host(:focus-within) {
        box-shadow: 0 0 0 2px var(--color-plain-a);
      }

      :host(.fullscreen) {
        position: fixed;
        top: 0;
        left: 0;
        z-index: 9;
        width: 100vw;
        height: 100vh;
        max-height: 100vh;
        border-radius: 0;
      }
    `,

    css`
      @use 'sass:list';

      .font-layer,
      .color-layer,
      .link-layer,
      .table-layer {
        visibility: hidden;
        position: absolute;
        left: 0;
        top: 0;
        z-index: 99;
        width: 80px;
        padding: 5px 0;
        line-height: 25px;
        background: #fff;
        box-shadow: 0 0 8px rgba(0, 0, 0, 0.2);
        font-size: 13px;
        opacity: 0;
        transition: all ease-in-out 0.2s;

        &.fadein {
          visibility: visible;
          top: 34px;
          opacity: 1;
        }
      }

      .font-layer {
        span {
          display: block;
          padding: 0 8px;

          &:hover {
            background: #f7f8fb;
          }
        }
      }
      $colors: #f3f5fb, #dae1e9, #62778d, #58d68d, #3fc2a7, #52a3de, #ac61ce,
        #ffb618, #e67e22, #ff5061, #ff0000, #000000;

      .color-layer {
        display: flex;
        flex-flow: row wrap;
        justify-content: center;
        left: 30px;
        width: 100px;

        span {
          width: 20px;
          height: 20px;
          margin: 2px;

          @for $i from 1 to 13 {
            &:nth-child(#{$i}) {
              background: list.nth($colors, $i);
            }
          }
        }
      }

      .table-layer {
        display: flex;
        flex-wrap: wrap;
        justify-content: space-between;
        left: 240px;
        width: 200px;
        height: 200px;
        padding: 2px;
        background: #fff;

        span {
          width: 20px;
          height: 20px;
          background: var(--color-plain-1);

          &.active {
            background: rgba(77, 182, 172, 0.3);
          }
        }
      }

      .link-layer {
        display: flex;
        flex-direction: column;
        left: 330px;
        width: 230px;
        padding: 8px;

        wc-button {
          width: 40px;
          margin-top: 8px;
        }
      }
    `
  ]

  #value = ''
  #gridx = 0
  #gridy = 0

  #select = null

  get value() {
    let html = this.$refs.editor?.innerHTML || ''

    html = html.replace(/<\!\-\-(.*?)\-\->/g, '')
    if (~html.indexOf('<table>') && !html.startsWith('<style>table')) {
      html =
        '<style>table{border-spacing:0;border-collapse:collapse;}table tr{background:#fff;}table thead tr{background:#f3f5fb;}table th,table td{padding:6px 12px;border:1px solid #dae1e9;}table th{font-weight: bold;}</style>' +
        html
    }
    return html
  }

  set value(val) {
    if (this.$refs.editor) {
      this.#value = val
      this.$refs.editor.innerHTML = val
      this.$emit('input')
    } else {
      nextTick(_ => (this.value = val))
    }
  }

  #updateStat() {
    if (this.$refs.editor) {
      if (this.readOnly || this.disabled) {
        this.$refs.editor.removeAttribute('contenteditable')
      } else {
        this.$refs.editor.setAttribute('contenteditable', '')
      }
    } else {
      nextTick(_ => this.#updateStat())
    }
  }

  #hideLayers() {
    this.$refs.font.classList.remove('fadein')
    this.$refs.color.classList.remove('fadein')
    this.$refs.link.classList.remove('fadein')
    this.$refs.table.classList.remove('fadein')
  }

  // 处理图片
  #handleImage(ev, file) {
    if (ev && ev.type === 'change') {
      file = ev.target.files[0]
      ev.target.value = ''
    }
    this.$emit('upload', {
      detail: {
        file,
        send: link => {
          this.$refs.editor.focus()
          this.restoreSelection()
          this.exec(ACTTION.image, link)
          this.saveSelection()

          //  修正插入的图片,宽度不得超出容器
          this.$refs.editor.querySelectorAll('img').forEach(_ => {
            _.style.maxWidth = '100%'
          })
        }
      }
    })
  }

  #handlePaste(ev) {
    let html = ev.clipboardData.getData('text/html')
    let txt = ev.clipboardData.getData('text/plain')
    let items = ev.clipboardData.items

    if (this.readOnly || this.disabled) {
      return
    }

    // 先文件判断, 避免右键单击复制图片时, 当成html处理
    if (items && items.length) {
      let blob = null
      for (let it of items) {
        if (it.type.indexOf('image') > -1) {
          blob = it.getAsFile()
        }
      }

      if (blob) {
        return this.#handleImage(null, blob)
      }
    }

    if (html) {
      html = html
        .replace(/\t/g, '  ')
        .replace(/<\/?(meta|link|script)[^>]*?>/g, '')
        .replace(/<!--[\w\W]*?-->/g, '')
        .replace(
          /<a[^>]*? href\s?=\s?["']?([^"']*)["']?[^>]*?>/g,
          '<a href="$1">'
        )
        .replace(
          /<img[^>]*? src\s?=\s?["']?([^"']*)["']?[^>]*?>/g,
          '<img src="$1">'
        )
        .replace(/<(?!a|img)([\w\-]+)[^>]*>/g, '<$1>')
        .replace(/<xml[^>]*?>[\w\W]*?<\/xml>/g, '')
        .replace(/<style>[\w\W]*?<\/style>/g, '')

      return this.exec('insertHtml', html)
    }

    if (txt) {
      return this.exec('insertText', txt)
    }
  }

  #toolbarClick(ev) {
    var target = ev.target
    var act

    this.restoreSelection()

    if (ev.target === ev.currentTarget) {
      return
    }

    if (this.readOnly || this.disabled) {
      return
    }

    while (target.tagName !== 'SPAN') {
      target = target.parentNode
    }

    act = target.dataset.act

    this.#hideLayers()

    switch (act) {
      case 'font':
      case 'color':
      case 'link':
      case 'table':
        this.$refs[act].classList.add('fadein')
        break

      case 'image':
        // 这里不作任何处理
        break

      case 'copy':
        navigator.clipboard.writeText(this.value)
        break

      case 'fullscreen':
        this.classList.toggle('fullscreen')
        break

      default:
        this.$refs.editor.focus()
        this.restoreSelection()
        this.exec(ACTTION[act])
        this.saveSelection()
        break
    }
  }

  #changeFontSize(ev) {
    if (ev.target === ev.currentTarget) {
      return
    }
    this.$refs.font.classList.remove('fadein')
    this.$refs.editor.focus()
    this.restoreSelection()
    this.exec(ACTTION.font, ev.target.dataset.size)
    this.saveSelection()
  }

  #chnageColor(ev) {
    if (ev.target === ev.currentTarget) {
      return
    }
    this.$refs.color.classList.remove('fadein')
    this.$refs.editor.focus()
    this.restoreSelection()
    this.exec(ACTTION.color, ev.target.dataset.value)
    this.saveSelection()
  }

  #insertLink(ev) {
    let value = this.$refs.linkinput.value.trim()
    if (value) {
      this.$refs.link.classList.remove('fadein')
      this.$refs.editor.focus()
      this.restoreSelection()
      this.exec(ACTTION.link, value)
      this.saveSelection()
      this.$refs.linkinput.value = ''
    }
  }

  #insertTable(ev) {
    let th = `<th>&nbsp;</th>`.repeat(this.#gridx + 1)
    let td = `<td>&nbsp;</td>`.repeat(this.#gridx + 1)
    this.exec(
      'insertHtml',
      `<br>
      <table>
      <thead><tr>${th}</tr></thead>
      <tbody>${`<tr>${td}</tr>`.repeat(this.#gridy + 1)}</tbody>
    </table><br>`
    )
    this.$refs.table.classList.remove('fadein')
  }

  #tableSelect(ev) {
    let grids = Array.from(this.$refs.table.children)
    if (ev.type === 'mousemove') {
      if (ev.target === ev.currentTarget) {
        return
      }
      let idx = +ev.target.dataset.idx
      let x = getX(idx)
      let y = getY(idx)
      // 避免每次遍历完所有的节点
      let max = Math.max(getIndex(this.#gridx, this.#gridy), idx) + 1

      if (x === this.#gridx && y === this.#gridy) {
        return
      }
      this.#gridx = x
      this.#gridy = y

      for (let i = 0; i < max; i++) {
        let _x = getX(i)
        let _y = getY(i)
        grids[i].classList.toggle('active', _x <= x && _y <= y)
      }
    } else {
      this.#gridx = -1
      this.#gridy = -1
      grids.forEach(it => it.classList.remove('active'))
    }
  }

  // 保存选中
  saveSelection() {
    var gs = this.root.getSelection()
    if (gs.getRangeAt && gs.rangeCount) {
      this.#select = gs.getRangeAt(0)
    }
  }

  //  清除选中并重置选中
  restoreSelection() {
    var gs = this.root.getSelection()
    if (this.#select) {
      try {
        gs.removeAllRanges()
      } catch (err) {}
      gs.addRange(this.#select)
    }
  }
  // 执行命令
  exec(cmd, val = '') {
    document.execCommand(cmd, false, val)
  }

  mounted() {
    this.exec('styleWithCSS', true)
    this._clickoutsideFn = outsideClick(this, _ => this.#hideLayers())
  }

  unmounted() {
    clearOutsideClick(this._clickoutsideFn)
  }

  render() {
    return html`
      <div class="editor">
        <section class="toolbar" @click=${this.#toolbarClick}>
          ${TOOLBAR.map(
            it => html`
              <span data-act=${it}>
                <svg class="icon" viewBox="0 0 1024 1024">
                  <path d=${ICONS[it]} />
                </svg>
                ${it === 'image'
                  ? html`<input
                      type="file"
                      :disabled=${this.readOnly || this.disabled}
                      @change=${this.#handleImage}
                      accept="image/*"
                    />`
                  : ''}
              </span>
            `
          )}
        </section>

        <div class="wrapper" ref="cont" @click=${this.#hideLayers}>
          <div
            ref="editor"
            contenteditable="true"
            class="typearea"
            @mouseleave=${this.saveSelection}
            @paste.prevent=${this.#handlePaste}
          ></div>
        </div>

        <div
          class="font-layer noselect"
          ref="font"
          @click=${this.#changeFontSize}
        >
          <span data-size="6">6号字体</span>
          <span data-size="5">5号字体</span>
          <span data-size="4">4号字体</span>
          <span data-size="3">3号字体</span>
          <span data-size="2">2号字体</span>
        </div>

        <div
          class="color-layer noselect"
          ref="color"
          @click=${this.#chnageColor}
        >
          ${COLORS.map(c => html`<span data-value=${c}></span>`)}
        </div>

        <div
          class="table-layer noselect"
          ref="table"
          @click=${this.#insertTable}
          @mousemove=${this.#tableSelect}
          @mouseleave=${this.#tableSelect}
        >
          ${range(81).map((_, n) => html`<span data-idx=${n}></span>`)}
        </div>

        <div class="link-layer noselect" ref="link">
          <wc-input ref="linkinput" placeholder="请输入链接地址"></wc-input>
          <wc-button size="m" @click=${this.#insertLink}>插入</wc-button>
        </div>
      </div>
    `
  }
}

Editor.reg('editor')
