/**
 * {按钮组件}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/03/06 15:17:25
 */

import { nextTick, css, html, Component, classMap, outsideClick } from 'wkit'
import '../icon/index.js'

const ANIMATION = {
  duration: 100,
  custom: [
    { transform: 'scaleY(0)', opacity: 0 },
    { transform: 'scaleY(1)', opacity: 1 }
  ]
}

class Input extends Component {
  static props = {
    readOnly: false,
    autofocus: false,
    disabled: false,
    clearable: false,
    icon: '',
    placeholder: '',
    maxlength: { type: Number, default: null },
    minlength: { type: Number, default: null },
    value: 'str!',
    lazy: 'num!0' // 并发拦截时间, 单位毫秒
  }
  #list = []
  #selectIndex = -1
  #listShowing = false
  #stamp = 0

  static styles = [
    css`
      :host {
        position: relative;
        display: inline-flex;
        min-width: 188px;
        height: 32px;
        user-select: none;
        -moz-user-select: none;
        color: var(--color-dark-1);
        border-radius: 3px;
        cursor: text;
        transition: box-shadow 0.15s linear;
      }

      .label {
        flex: 1;
        display: flex;
        justify-content: center;
        align-items: center;
        height: 100%;
        font-size: 14px;
        border: 1px solid var(--wc-input-border-color, var(--color-grey-2));
        border-radius: inherit;
        background: var(--bg-color, #fff);
        color: inherit;
        cursor: inherit;

        input {
          flex: 1;
          min-width: 36px;
          width: 0;
          height: 100%;
          padding: 0 8px;
          line-height: 1;
          border: 0;
          border-radius: inherit;
          font: inherit;
          color: inherit;
          background: none;
          outline: none;
          box-shadow: none;
          cursor: inherit;

          &::placeholder {
            color: var(--color-grey-1);
          }
        }

        /* ----- */
        .close {
          --wc-icon-size: 18px;
          margin: 0 8px 0 4px;
          padding: 4px;
          border-radius: 50%;
          color: var(--color-grey-2);
          cursor: pointer;
          transition: background 0.15s linear;

          &:hover {
            background: var(--color-plain-1);
          }
        }
        .icon {
          --wc-icon-size: 16px;
          margin: 0 8px 0 4px;
          color: var(--color-grey-2);
        }
      }

      .suggestion {
        overflow: hidden;
        position: absolute;
        z-index: 1;
        left: 0;
        top: calc(100% + 4px);
        width: 100%;
        padding: 4px 0;
        border-radius: 4px;
        box-shadow: 0 2px 5px rgba(0, 0, 0, 0.15);
        transform-origin: top;

        .list {
          width: 100%;
          background: #fff;
        }
        &.hide {
          display: none;
        }

        li {
          overflow: hidden;
          width: 100%;
          height: 30px;
          line-height: 30px;
          padding: 0 8px;
          text-overflow: ellipsis;
          white-space: nowrap;
          cursor: pointer;

          &:hover,
          &[focus] {
            background: var(--color-plain-2);
          }
        }
      }

      /* 额外样式 */
      :host([round]) {
        border-radius: 26px;

        .label input {
          padding: 0 10px;
        }
      }

      :host(:focus-within) {
        box-shadow: 0 0 0 2px var(--color-plain-a);
      }
      /* --- */
      :host([disabled]) {
        pointer-events: none;
        cursor: not-allowed;

        .label {
          border-color: var(--color-grey-1);
          background: var(--color-plain-1);
          opacity: 0.6;
        }
      }
      :host([readonly]) {
        cursor: default;
      }
    `,
    //尺寸
    css`
      @use 'sass:map';
      $sizes: (
        m: (
          w: 128px,
          h: 24px,
          f: 12px
        ),
        // l default
        xl:
          (
            w: 224px,
            h: 36px,
            f: 14px
          ),
        xxl: (
          w: 288px,
          h: 44px,
          f: 14px
        )
      );

      @loop $s, $v in $sizes {
        :host([size='#{$s}']) {
          min-width: map.get($v, 'w');
          height: map.get($v, 'h');
          font-size: map.get($v, 'f');
          .label {
            height: map.get($v, 'h');
            font-size: map.get($v, 'f');
          }
          .icon {
            --wc-icon-size: #{map.get($v, 'f')};
          }
        }
        :host([size='#{$s}'][circle]) {
          width: map.get($v, 'h');
          height: map.get($v, 'h');
        }
      }
    `
  ]
  renderClear() {
    return html`<wc-icon class="close" name="close" @click=${this.clear} />`
  }

  render() {
    let classes = classMap({
      suggestion: true,
      hide: !this.#list.length
    })

    return html`
      <div class="label">
        <slot name="prepend">
          <wc-icon class="icon" name=${this.icon}></wc-icon>
        </slot>
        <input
          ref="input"
          @input=${this.handleInput}
          @change=${this.handleChange}
          @keydown=${this.handleKeyDown}
          @focus=${this.handleFocus}
          placeholder=${this.placeholder}
          maxlength=${this.maxlength}
          minlength=${this.minlength}
          disabled=${this.disabled}
          readonly=${this.readOnly}
          autofocus=${this.autofocus}
          :value=${this.value}
        />
        ${this.clearable && this.value ? this.renderClear() : ''}
        <slot name="append"></slot>
        <div class=${classes} ref="suggestion" #animation=${ANIMATION}>
          <div ref="scroller" class="scroller">
            <ul class="list" @click=${this.handleClickItem} ref="list">
              ${this.#list.map(
                (li, idx) =>
                  html`<li
                    focus=${this.#selectIndex === idx ? true : null}
                    index=${idx}
                  >
                    ${li.value}
                  </li>`
              )}
            </ul>
          </div>
        </div>
      </div>
    `
  }

  handleInput(e) {
    let { lazy } = this
    this.value = e.currentTarget.value

    if (lazy && Date.now() - this.#stamp < lazy) {
      return
    }
    this.#stamp = Date.now()

    this.emitFetchSuggest()
  }
  handleClickItem(e) {
    let index = e.target.getAttribute('index')
    this.#selectIndex = index
    this.emitSelect()
  }
  clear() {
    this.$refs.input.value = ''
    this.value = ''
    this.$emit('change')
    this.$emit('input')
  }
  handleChange() {
    this.$emit('change')
  }
  handleKeyDown(e) {
    let { lazy, minlength, value } = this

    if (e.keyCode === 13) {
      e.preventDefault()
      if (this.#selectIndex > -1 && this.#listShowing) {
        return this.emitSelect()
      }
      if (lazy && Date.now() - this.#stamp < lazy) {
        return
      }
      this.#stamp = Date.now()
      if (minlength && value.length < minlength) {
        return
      }
      return this.$emit('submit')
    }

    if (e.keyCode === 38 || e.keyCode === 40) {
      e.preventDefault()
      let step = e.keyCode === 38 ? -1 : 1

      this.#selectIndex += step
      if (this.#selectIndex < 0) {
        this.#selectIndex = 0
      }
      if (this.#selectIndex > this.#list.length - 1) {
        this.#selectIndex = this.#list.length - 1
      }

      this.$requestUpdate()
    }
  }
  // 触发列表选择
  emitSelect() {
    let item = this.#list[this.#selectIndex]
    this.value = item.value

    this.$refs.suggestion.$animate(true)
    this.#listShowing = false
    this.$requestUpdate()

    this.$emit('change')
    this.$emit('input')
    this.$emit('select', {
      index: this.#selectIndex,
      value: item
    })
  }
  emitFetchSuggest() {
    this.$emit('fetch-suggest', {
      value: this.value,
      send: list => {
        this.#list = list.slice(0, 10)
        this.#selectIndex = -1
        this.$requestUpdate()
      }
    })
  }
  handleFocus() {
    if (!this.#listShowing) {
      this.#listShowing = true
      this.$refs.suggestion.$animate()
    }
  }
  mounted() {
    if (this.autofocus) {
      // 火狐浏览器需要手动focus()才能聚焦成功
      nextTick(_ => this.$refs.input.focus())
    }
    outsideClick(this, () => {
      this.#listShowing = false
      this.$refs.suggestion.$animate(true)
    })
  }
}

Input.reg('input')
