/**
 * markdown解析器
 * @author yutent<yutent.io@gmail.com>
 * @date 2020/02/07 17:14:19
 */

const HR_LIST = ['=', '-', '_', '*']
const LIST_RE = /^(([\+\-\*])|(\d+\.))\s/
const TODO_RE = /^[\+\-\*]\s\[(x|\s)\]\s/
const ESCAPE_RE = /\\([-+*_`\]\[\(\)])/g
const QLINK_RE = /^\[(\d+)\]: ([\S]+)\s*?((['"])[\s\S]*?\4)?\s*?$/
const TAG_RE = /<([\w\-]+)([\w\W]*?)>/g
const ATTR_RE = /\s*?on[a-zA-Z]+="[^"]*?"\s*?/g
const CODEBLOCK_RE = /```(.*?)([\w\W]*?)```/g
const BLOCK_RE = /<([\w\-]+)([^>]*?)>([\w\W]*?)<\/\1>/g
const STYLE_RE = /<style[^>]*?>([\w\W]*?)<\/style>/g
const SINGLE_LINE_HTML_RE = /^\s*?<([\w\-]+)[^>]*?>.*?<\/\1>\s*?$/
const MULTI_LINE_HTML_RE = /^\s*?<\/?[^>]*?>/

const INLINE = {
  code: /`([^`]*?[^`\\\s])`/g,
  strong: [/__([\s\S]*?[^\s\\])__(?!_)/g, /\*\*([\s\S]*?[^\s\\])\*\*(?!\*)/g],
  em: [/_([\s\S]*?[^\s\\])_(?!_)/g, /\*([\s\S]*?[^\s\\*])\*(?!\*)/g],
  del: /~~([\s\S]*?[^\s\\~])~~/g,
  qlink: /\[([^\]]*?)\]\[(\d*?)\]/g, // 引用链接
  img: /\!\[([^\]]*?)\]\(([^)]*?)\)/g,
  a: /\[([^\]]*?)\]\(([^)]*?)(\s+"([\s\S]*?)")*?\)/g,
  qlist: /((<blockquote class="md\-quote">)*?)([\+\-\*]|\d+\.) (.*)/ // 引用中的列表
}

const Helper = {
  // 是否分割线
  isHr(str) {
    var s = str[0]
    if (HR_LIST.includes(s)) {
      return str.trim() === s.repeat(3)
    }
    return false
  },
  // 是否列表, -1不是, 1为有序列表, 0为无序列表
  isList(str) {
    var v = str.trim()
    if (LIST_RE.test(v)) {
      var n = +v[0]
      if (n === n) {
        return 1
      } else {
        return 0
      }
    }
    return -1
  },
  // 是否任务列表
  isTodo(str) {
    var v = str.trim()
    if (TODO_RE.test(v)) {
      return v[3] === 'x' ? 1 : 0
    }
    return -1
  },
  ltrim(str) {
    if (str.trimStart) {
      return str.trimStart()
    }
    return str.replace(/^\s+/, '')
  },
  isQLink(str) {
    if (QLINK_RE.test(str)) {
      // l: link,  t: title, $1: index
      return { [RegExp.$1]: { l: RegExp.$2, t: RegExp.$3 } }
    }
    return false
  },
  isTable(str) {
    return /^\|.+?\|$/.test(str)
  }
}

const Decoder = {
  // 内联样式
  inline(str) {
    return str
      .replace(INLINE.code, (m, str) => {
        str = str.replace(/([_*~])/g, '\\$1')
        return `<code class="inline">${str}</code>`
      })
      .replace(INLINE.strong[0], '<strong>$1</strong>')
      .replace(INLINE.strong[1], '<strong>$1</strong>')
      .replace(INLINE.em[0], '<em>$1</em>')
      .replace(INLINE.em[1], '<em>$1</em>')
      .replace(INLINE.del, '<del>$1</del>')
      .replace(INLINE.img, '<img src="$2" alt="$1">')
      .replace(INLINE.a, (m1, txt, link, m2, attr = '') => {
        var tmp = attr
          .split(';')
          .filter(_ => _)
          .map(_ => {
            var a = _.split('=')
            if (a.length > 1) {
              return `${a[0]}="${a[1]}"`
            } else {
              return `title="${_}"`
            }
          })
          .join(' ')

        return `<a href="${link.trim()}" ${tmp}>${txt}</a>`
      })
      .replace(INLINE.qlink, (m, txt, n) => {
        var _ = this.__LINKS__[n]
        if (_) {
          var a = _.t ? `title=${_.t}` : ''
          return `<a href="${_.l}" ${a}>${txt}</a>`
        } else {
          return m
        }
      })
      .replace(ESCAPE_RE, '$1') // 处理转义字符
  },
  // 分割线
  hr() {
    return `\n\n<hr>\n\n`
  },
  // 标题
  head(str) {
    if (str.startsWith('#')) {
      return str.replace(/^(#{1,6}) (.*)/, (p, m1, m2) => {
        m2 = m2.trim()
        let level = m1.trim().length
        let hash = m2.replace(/\s/g, '').replace(/<\/?[^>]*?>/g, '')

        if (level === 1) {
          return `<h1>${m2}</h1>`
        } else {
          return `\n\n<h${level}><a href="#${hash}" id="${hash}" class="md-head-link">${m2}</a></h${level}>\n`
        }
      })
    }
    return false
  },

  // 任务
  task(str) {
    var todoChecked = Helper.isTodo(str)
    if (~todoChecked) {
      var word = str.replace(TODO_RE, '').trim()
      var stat = todoChecked === 1 ? 'checked' : ''
      var txt = todoChecked === 1 ? `<del>${word}</del>` : word

      return `\n<section><wc-checkbox readonly ${stat}>${txt}</wc-checkbox></section>`
    }
    return false
  }
}

function fixed(str) {
  // 去掉\r, 将\t转为空格(2个)
  return str
    .replace(/\r\n|\r/g, '\n')
    .replace(/\t/g, '  ')
    .replace(/\u00a0/g, ' ')
    .replace(/\u2424/g, '\n')
}

function trimBr(str) {
  return str.endsWith('<br>') ? str.slice(0, -4) : str
}

class Tool {
  constructor(list, links) {
    this.list = list
    this.__LINKS__ = links
  }

  // 初始化字符串, 处理多余换行等
  static init(str = '') {
    let links = {}
    let list = []
    let lines = fixed(str).split('\n')
    let isCodeBlock = false // 是否代码块
    let isTable = false // 是否表格
    let emptyLines = 0 //连续空行的数量

    for (let it of lines) {
      let tmp = it.trim()

      // 非空行
      if (tmp) {
        emptyLines = 0
        if (tmp.startsWith('```')) {
          if (isCodeBlock) {
            list.push('\n</xmp></wc-code>\n')
          } else {
            list.push(
              tmp.replace(/^```([\w\#\-]*?)$/, `\n<wc-code lang="$1"><xmp>`)
            )
          }
          isCodeBlock = !isCodeBlock
        } else {
          var qlink
          if (isCodeBlock) {
            it = it
              .replace('\\`\\`\\`', '```')
              .replace(/<(\/?)xmp>/g, '&lt;$1xmp&gt;') // 转义这个特殊标签
          } else {
            if (Helper.isTable(tmp) && !isTable) {
              var thead = tmp.split('|')
              // 去头去尾
              thead.shift()
              thead.pop()
              list.push(
                `\n<table>\n<thead><tr>${thead
                  .map(_ => `<th>${_}</th>`)
                  .join('')}</tr></thead>\n<tbody>`
              )
              isTable = true
              continue
            } else {
              it = it
                // 非代码块进行xss过滤
                .replace(INLINE.code, (m, txt) => {
                  return `\`${txt
                    .replace(/</g, '&lt;')
                    .replace(/>/g, '&gt;')}\``
                })
                .replace(/<(\/?)script[^>]*?>/g, '&lt;$1script&gt;')
                .replace(TAG_RE, (m, name, attr = '') => {
                  // 过滤所有onXX=""事件属性
                  attr = attr.replace(ATTR_RE, ' ').trim()
                  if (attr) {
                    attr = ' ' + attr
                  }
                  return `<${name}${attr}>`
                })
              // 不在代码块中, 才判断引用声明
              qlink = Helper.isQLink(it)
            }
          }

          if (qlink) {
            Object.assign(links, qlink)
          } else {
            list.push(it)
          }
        }
      } else {
        if (list.length === 0) {
          continue
        }
        if (isTable) {
          isTable = false
          list.push('</tbody>\n</table>\n')
          continue
        }
        if (isCodeBlock || emptyLines < 2) {
          emptyLines++
          list.push(tmp)
        }
      }
    }
    return new this(list, links)
  }

  parse() {
    let html = ''
    let isCodeBlock = false // 是否代码块
    let emptyLines = 0 //连续空行的数量
    let isBlockquote = false
    let isTable = false
    let tableAlign = null
    let blockquoteLevel = 0
    let isParagraph = false

    let isQList = false
    let qListTag = null
    let isList = false
    let listTagQueue = []

    let isQuoteList = false // 引用中的列表, 只支持一层级
    let quoteListStyle = 0 // 1有序,  2 无序
    let isHtmlBlock = false // 是否原生html代码块

    //
    for (let it of this.list) {
      // 非空行
      if (it) {
        // 缓存上一个连续空行数
        let lastEmptyLines = emptyLines

        emptyLines = 0

        if (~it.indexOf('<table>') || ~it.indexOf('</table>')) {
          if (isList) {
            if (isTable) {
              html += it + `${'  '.repeat(listTagQueue.length)}</li>\n`
            } else {
              html = html.replace(/<\/li>\n*?$/, '') + it
            }
          } else {
            html += it
          }
          isTable = !isTable
          tableAlign = true
          isHtmlBlock = false
          continue
        }

        if (isTable) {
          let tmp = it.split('|').map(_ => _.trim())
          tmp.shift()
          tmp.pop()

          // 表格分割行, 配置对齐方式的
          if (tableAlign === true) {
            tableAlign = tmp.map(a => {
              a = a.split(/\-+/)
              if (a[0] === ':' && a[1] === ':') {
                return 'align="center"'
              }
              if (a[1] === ':') {
                return 'align="right"'
              }
              return ''
            })
            continue
          }
          html += `<tr>${tmp
            .map(
              (_, i) =>
                `<td ${tableAlign[i]}>${Decoder.inline.call(this, _)}</td>`
            )
            .join('')}</tr>`
          continue
        }

        // wc-code标签直接拼接, 判断时多拼一个 < 和 >,
        // 是为了避免在 wc-markd嵌入代码块示例时, 将其内容编译为html
        if (it.includes('<wc-code') || it.includes('wc-code>')) {
          // 如果之前未闭合的段落, 先闭合
          if (isParagraph) {
            isParagraph = false
            html += '</p>\n'
          }
          if (isList) {
            if (isCodeBlock) {
              html += it + `${'  '.repeat(listTagQueue.length)}</li>\n`
            } else {
              html = html.replace(/<\/li>\n*?$/, '') + it
            }
          } else {
            html += it
          }
          isCodeBlock = !isCodeBlock
          isHtmlBlock = false
          continue
        }

        // 同上代码块的处理
        if (isCodeBlock) {
          html += '\n' + it
          continue
        }

        // 无属性标签
        if (Helper.isHr(it)) {
          html += Decoder.hr()
          continue
        }

        // 处理常规样式之前, 先判断是否为原生html文本
        // 避免处理完常规样式之后生成的html干扰结果
        let isSingleLineHtml = SINGLE_LINE_HTML_RE.test(it)
        let isMultiLineHtml = !isSingleLineHtml && MULTI_LINE_HTML_RE.test(it)

        // 优先处理一些常规样式
        it = Decoder.inline.call(this, it)

        // 标题只能是单行

        let head = Decoder.head(it)
        if (head) {
          html = trimBr(html)
          // 有段落先结束
          if (isParagraph) {
            isParagraph = false
            html += '</p>'
          }
          // 有列表也先结束
          if (isList) {
            html = this.#closeList(html, listTagQueue, isParagraph)
            isList = false
          }
          isHtmlBlock = false
          html += head
          continue
        }

        // 引用
        if (it.startsWith('>')) {
          let innerQuote // 是否有缩进引用
          it = it.replace(/^(>+) /, (p, m) => {
            let len = m.length
            let tmp = ''
            let loop = len
            // 若之前已经有一个未闭合的引用, 需要减去已有缩进级别, 避免产生新的引用标签
            if (isBlockquote) {
              loop = len - blockquoteLevel
            }

            while (loop > 0) {
              loop--
              tmp += `\n<blockquote class="md-quote">`
            }

            blockquoteLevel = len
            innerQuote = !!tmp
            return tmp
          })

          if (isParagraph) {
            isParagraph = false
            html = trimBr(html)
            html += '</p>\n'
          }

          if (isBlockquote) {
            // 没有新的缩进引用时, 才添加换行
            if (innerQuote) {
              // 之前有引用的列表时, 直接结束列表
              if (isQuoteList) {
                html += `</${qListTag}>\n`
                isQuoteList = false
              }
            }
          }

          let qListChecked = it.match(INLINE.qlist)
          if (qListChecked) {
            let tmp1 = qListChecked[1] // 缩进的标签
            let tmp2 = +qListChecked[3] // 有序还是无序
            let tmp3 = qListChecked.pop() // 文本
            let currListStyle = tmp2 === tmp2 ? 1 : 2
            var qlist = ''

            // 已有列表
            // 因为只支持一层级的列表, 所以同一级别不区分有序无序, 强制统一
            if (isQuoteList === false) {
              isQuoteList = true
              if (currListStyle === 1) {
                qListTag = 'ol'
              } else {
                qListTag = 'ul'
              }
              qlist += `\n  <${qListTag}>`
            }

            quoteListStyle = currListStyle

            qlist += `<li>${tmp3}</li>`
            html += tmp1 + qlist
            isQList = true
          } else {
            if (innerQuote === false) {
              html += '<br>'
            }
            html += it
          }

          isHtmlBlock = false
          isParagraph = false
          isBlockquote = true
          continue
        }

        // 任务
        let task = Decoder.task(it)
        if (task) {
          html += task
          isHtmlBlock = false
          continue
        }

        // 列表
        let listChecked = Helper.isList(it)
        if (~listChecked) {
          // 左侧空格长度
          let tmp = Helper.ltrim(it)
          let ltrim = it.length - tmp.length
          let word = tmp.replace(LIST_RE, '').trim()
          let _level = listTagQueue.length - 1 // 上一个缩进级别
          let level = Math.floor(ltrim / 2) // 当前缩进级别
          let tag = listChecked === 1 ? 'ol' : 'ul'

          if (isParagraph) {
            isParagraph = false
            html = trimBr(html)
            html += '</p>\n'
          }

          if (isList) {
            // 大于上一个缩进
            if (level > _level) {
              listTagQueue.push(tag)
              // 需要先把上一个列表的结束符去掉
              html = html.trim().replace(/<\/li>$/, '')
              html += `\n${'  '.repeat(level + 1)}<${tag}>\n${'  '.repeat(
                level + 1
              )}<li>${word}</li>\n`
            }
            // 跟上一个平级
            else if (level === _level) {
              html += `${'  '.repeat(level + 1)}<li>${word}</li>\n`
            }
            // 比上一个缩进要小
            else {
              for (let i = _level - level; i > 0; i--) {
                let _tag = listTagQueue.pop()

                html += `${'  '.repeat(level + i + 1)}</${_tag}>\n${'  '.repeat(
                  level + i
                )}</li>\n`
              }

              html += `${'  '.repeat(level + 1)}<li>${word}</li>\n`
            }
          } else {
            listTagQueue.push(tag)
            html += `<${tag}>\n  <li>${word}</li>\n`
          }

          isList = true
          isHtmlBlock = false
          continue
        }

        // 无"> "前缀的引用, 继续拼到之前的, 并且不换行
        if (isBlockquote) {
          if (isQList) {
            html = html.replace(/<\/li>\n*?$/, '') + ` ${it}</li>`
          } else {
            html += ' ' + it
          }
          continue
        }

        if (isHtmlBlock || isSingleLineHtml) {
          if (isParagraph) {
            isParagraph = false
            html = trimBr(html)
            html += '</p>'
          }
          html += `${it}\n`
        } else if (isMultiLineHtml) {
          if (isParagraph) {
            isParagraph = false
            html = trimBr(html)
            html += '</p>'
          }
          isHtmlBlock = !isHtmlBlock
          html += `${it}\n`
        } else {
          if (isList) {
            if (isParagraph) {
              if (lastEmptyLines > 0) {
                html = trimBr(html)
                html += `<br>${it}`
                continue
              } else {
                html += ` ${it}`
              }
            } else {
              if (lastEmptyLines > 0) {
                html = html.replace(/<\/li>\n*?$/, '') + `<p>${it}`
              } else {
                html = html.replace(/<\/li>\n*?$/, '') + ` ${it}</li>\n`
                continue
              }
            }
          } else {
            if (isParagraph) {
              html += ` ${it}\n`
            } else {
              html += `\n<p>${it}`
            }
          }
          isParagraph = true
        }
      } else {
        // 如果是在代码中, 直接拼接, 并加上换行
        if (isCodeBlock) {
          html += it + '\n'
        } else {
          emptyLines++
          //
          // 引用结束
          if (isBlockquote) {
            if (isQList) {
              html += `\n</${qListTag}>\n`
              isQList = false
            }
            html = this.#closeQuotes(html, blockquoteLevel)
            isBlockquote = false
            blockquoteLevel = 0
          }

          if (isList && emptyLines > 1) {
            html = this.#closeList(html, listTagQueue, isParagraph)
            isList = false
            emptyLines = 0
            if (isParagraph) {
              isParagraph = false
            }
          }

          if (isParagraph) {
            if (emptyLines > 1) {
              emptyLines = 0
              isParagraph = false
              html = trimBr(html)
              html += '</p>\n'
            } else {
              html += '<br>'
            }
          }
        }
      }
    }

    // 引用结束
    if (isBlockquote) {
      html = this.#closeQuotes(html, blockquoteLevel)
      blockquoteLevel = 0
    }

    if (isList) {
      html = this.#closeList(html, listTagQueue, isParagraph)
      if (isParagraph) {
        isParagraph = false
      }
    }

    if (isParagraph) {
      html += '</p>'
    }
    delete this.list
    delete this.__LINKS__
    return html.trim()
  }

  #closeList(html, tags, isParagraph) {
    if (isParagraph) {
      html = trimBr(html)
      html += '</p>\n'
    }
    while (tags.length) {
      let tag = tags.pop()
      html += `${'  '.repeat(tags.length ? tags.length + 1 : 0)}</${tag}>\n`
      if (tags.length) {
        html += `${'  '.repeat(tags.length)}</li>\n`
      }
    }
    return html
  }

  #closeQuotes(html, level) {
    while (level > 0) {
      level--
      html += '</blockquote>\n'
    }
    return html
  }
}

export default function (str) {
  return Tool.init(str).parse()
}
