/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/09/14 16:49:15
 */

import {
  css,
  html,
  Component,
  range,
  nextTick,
  classMap,
  outsideClick,
  clearOutsideClick
} from 'wkit'

import '../markd/index.js'
import '../form/input.js'
import '../form/button.js'

import { DEFAULT_TOOLS, html2md, Addon } from './helper.js'
import ICONS from './svg.js'

// 获取一维数组转二维的行
function getY(i) {
  return (i / 9) >> 0
}
//获取一维数组转二维的列
function getX(i) {
  return i % 9
}

function getIndex(x, y) {
  return x + y * 9
}

class MEditor extends Component {
  static watches = ['value']

  static props = {
    toolbar: {
      type: String,
      default: null,
      attribute: false,
      observer(v) {
        if (v === null) {
          this.#toolbar = [...DEFAULT_TOOLS]
        } else {
          this.#toolbar = v
            .split(',')
            .filter(it => it)
            .map(it => it.trim())
        }
      }
    },
    placeholder: '',
    readonly: false,
    disabled: false,
    autofocus: false
  }

  static styles = [
    css`
      :host {
        display: flex;
        min-width: 200px;
        max-height: 720px;
        min-height: 64px;
        border-radius: 3px;
        transition: box-shadow 0.15s linear;
        background: var(--wc-meditor-background, #fff);
      }
      .noselect {
        -webkit-touch-callout: none;
        -webkit-user-select: none;
        -moz-user-select: none;
        user-select: none;
      }
    `,

    css`
      .meditor {
        position: relative;
        flex: 1;
        display: flex;
        flex-direction: column;
        width: 100%;
        border: 1px solid var(--wc-editor-border-color, var(--color-grey-2));
        border-radius: inherit;
        font-size: 14px;
        background: #fff;
      }

      .toolbar {
        display: none;
        width: 100%;
        height: 34px;
        padding: 5px;
        line-height: 24px;
        border-bottom: 1px solid var(--color-grey-1);

        span {
          position: relative;
          overflow: hidden;
          display: flex;
          justify-content: center;
          align-items: center;
          width: 24px;
          height: 24px;
          margin: 0 3px;
          border-radius: 3px;
          color: var(--color-grey-3);

          .icon {
            overflow: hidden;
            width: 70%;
            height: 70%;
            fill: currentColor;
            color: #62778d;
          }
          input {
            position: absolute;
            width: 100%;
            height: 100%;
            opacity: 0;
          }

          &:hover,
          &.active {
            background: var(--color-plain-1);
          }
          &.active {
            color: var(--color-teal-1);
          }
        }

        &.active {
          display: flex;
        }
      }
    `,
    css`
      .editor-outbox {
        overflow: hidden;
        flex: 1;
        display: flex;
        width: 100%;
        border-radius: 3px;

        .editor,
        .preview {
          flex: 1;
          flex-shrink: 0;
        }

        .editor {
          height: 100%;
          padding: 5px 8px;
          line-height: 1.5;
          border: 0;
          font-size: var(--wc-meditor-font-size, 14px);
          font-family: Menlo, Monaco, Consolas, 'Courier New', monospace;
          color: var(--color-dark-1);
          background: none;
          outline: none;
          resize: none;
          cursor: inherit;

          &::placeholder {
            color: var(--color-grey-1);
          }
        }

        .preview {
          overflow: hidden;
          overflow-y: auto;
          display: none;
          padding: 6px 12px;
          border-left: 1px solid var(--color-plain-2);

          &.active {
            display: block;
          }
        }
      }
    `,
    css`
      :host([readonly]) {
        cursor: default;
        opacity: 0.8;
      }
      :host([disabled]) {
        cursor: not-allowed;
        opacity: 0.6;
      }
      :host([readonly]),
      :host([disabled]) {
        .toolbar {
          span:hover {
            background: none;
          }
        }
      }

      :host(:focus-within) {
        box-shadow: 0 0 0 2px var(--color-plain-a);
      }

      :host(.fullscreen) {
        position: fixed;
        top: 0;
        left: 0;
        z-index: 9;
        width: 100vw;
        height: 100vh;
        max-height: 100vh;
        border-radius: 0;
      }
    `,

    css`
      .font-layer,
      .link-layer,
      .table-layer {
        visibility: hidden;
        position: absolute;
        left: 0;
        top: 0;
        z-index: 99;
        height: 0;
        background: #fff;
        opacity: 0;
        box-shadow: 0 0 8px rgba(0, 0, 0, 0.2);
        transition: all ease-in-out 0.2s;

        &.fadein {
          visibility: visible;
          top: 34px;
          height: auto;
          opacity: 1;
        }
      }

      .font-layer {
        width: 120px;
        padding: 5px 0;
        font-family: Menlo, Monaco, Consolas, 'Courier New', monospace;

        span {
          display: block;
          padding: 0 8px;

          &:hover {
            background: #f7f8fb;
          }

          &:first-child {
            font-size: 24px;
          }
          &:nth-child(2) {
            font-size: 22px;
          }
          &:nth-child(3) {
            font-size: 20px;
          }
          &:nth-child(4) {
            font-size: 18px;
          }
          &:nth-child(5) {
            font-size: 16px;
          }
          &:nth-child(6) {
            font-size: 14px;
          }
        }
      }

      .table-layer {
        overflow: hidden;
        display: flex;
        flex-wrap: wrap;
        justify-content: space-between;
        left: 240px;
        width: 200px;
        padding: 2px;
        background: #fff;

        span {
          width: 20px;
          height: 20px;
          background: var(--color-plain-1);

          &.active {
            background: rgba(77, 182, 172, 0.3);
          }
        }

        &.fadein {
          height: 200px;
        }
      }

      .link-layer {
        display: flex;
        flex-direction: column;
        left: 330px;
        width: 230px;
        padding: 8px;

        wc-button {
          width: 40px;
          margin-top: 8px;
        }
      }
    `,

    css`
      .editor,
      wc-markd {
        &::-webkit-scrollbar {
          width: 6px;
        }

        &::-webkit-scrollbar-thumb {
          visibility: hidden;
          border-radius: 3px;
        }

        &:hover::-webkit-scrollbar-thumb {
          visibility: visible;
          background: rgba(0, 0, 0, 0.3);
        }
      }
    `
  ]

  get value() {
    try {
      return this.$refs.editor.value
    } catch (err) {
      return ''
    }
  }

  set value(val) {
    if (this.$refs.editor) {
      if (this.$refs.editor.value === val) {
        return
      }
      this.$refs.editor.value = val
      this.#updatePreview()
      this.$emit('input')
    } else {
      nextTick(_ => (this.value = val))
    }
  }

  #toolbar = [...DEFAULT_TOOLS]
  #value = ''

  #gridx = 0
  #gridy = 0

  previewEnabled = false

  #hideLayers() {
    this.$refs.header?.classList.remove('fadein')
    this.$refs.link?.classList.remove('fadein')
    this.$refs.table?.classList.remove('fadein')
  }

  // 处理图片
  #handleFile(ev, file, t = '!') {
    if (ev && ev.type === 'change') {
      file = ev.target.files[0]
      ev.target.value = ''
      t = ev.target.dataset.type === 'image' ? '!' : ''
    }
    this.$emit('upload', {
      detail: {
        file,
        send: link => {
          this.$refs.editor.focus()
          this.insert(`${t}[${file.name}](${link})`)
        }
      }
    })
  }

  #handlePaste(ev) {
    let html = ev.clipboardData.getData('text/html')
    let txt = ev.clipboardData.getData('text/plain')
    let items = ev.clipboardData.items

    if (this.readOnly || this.disabled) {
      return
    }

    // 先文件判断, 避免右键单击复制图片时, 当成html处理
    if (items && items.length) {
      let file = null

      for (let it of items) {
        file = it.getAsFile()
        if (file) {
          break
        }
      }

      if (file) {
        return this.#handleFile(
          null,
          file,
          file.type.includes('image') ? '!' : ''
        )
      }
    }

    if (html) {
      this.insert(html2md(html))
    } else if (txt) {
      this.insert(txt)
    }
  }

  #updatePreview() {
    if (this.previewEnabled) {
      this.$refs.view.code = this.value
    }
  }

  #toolbarClick(ev) {
    var elem = ev.target
    var act

    if (elem === ev.currentTarget) {
      return
    }

    if (this.readOnly || this.disabled) {
      return
    }

    while (elem.tagName !== 'SPAN') {
      elem = elem.parentNode
    }

    act = elem.dataset.act

    this.#hideLayers()

    if (act === 'image' || act === 'attach') {
      return
    }

    Addon[act].call(this, elem)
  }

  #setHeaderLevel(ev) {
    if (ev.target === ev.currentTarget) {
      return
    }
    this.$refs.header.classList.remove('fadein')
    this.$refs.editor.focus()
    let level = +ev.target.dataset.value
    Addon.h.call(this, level)
  }

  #insertLink(ev) {
    let value = this.$refs.linkinput.value.trim()
    if (value) {
      this.$refs.link.classList.remove('fadein')
      this.$refs.editor.focus()
      this.insert(`[${value}](${value}) `)
      this.$refs.linkinput.value = ''
    }
  }

  #insertTable(ev) {
    this.$refs.table.classList.remove('fadein')
    if (this.#gridx < 0 || this.#gridy < 0) {
      return
    }
    let thead = `\n\n${'| 表头 '.repeat(this.#gridx + 1)}|\n`
    let pipe = `${'| -- '.repeat(this.#gridx + 1)}|\n`
    let tbody = ('|    '.repeat(this.#gridx + 1) + '|\n').repeat(
      this.#gridy + 1
    )
    this.#gridx = -1
    this.#gridy = -1
    this.insert(thead + pipe + tbody, false)
  }

  #tableSelect(ev) {
    let grids = Array.from(this.$refs.table.children)
    if (ev.type === 'mousemove') {
      if (ev.target === ev.currentTarget) {
        return
      }
      let idx = +ev.target.dataset.idx
      let x = getX(idx)
      let y = getY(idx)
      // 避免每次遍历完所有的节点
      let max = Math.max(getIndex(this.#gridx, this.#gridy), idx) + 1

      if (x === this.#gridx && y === this.#gridy) {
        return
      }
      this.#gridx = x
      this.#gridy = y

      for (let i = 0; i < max; i++) {
        let _x = getX(i)
        let _y = getY(i)
        grids[i].classList.toggle('active', _x <= x && _y <= y)
      }
    } else {
      this.#gridx = -1
      this.#gridy = -1
      grids.forEach(it => it.classList.remove('active'))
    }
  }

  focus() {
    this.$refs.editor?.focus()
  }

  /**
   * 往文本框中插入内容
   * @param {String} val  [要插入的文本]
   * @param {Boolean} isSelect  [插入之后是否要选中文本]
   * @param {Boolean} offset  [选中文本时的偏移量]
   */
  insert(val = '', isSelect = false, offset = 0) {
    let $el = this.$refs.editor
    let start = $el.selectionStart
    let end = $el.selectionEnd
    let scrollTop = $el.scrollTop

    if (start || start === 0) {
      $el.value = $el.value.slice(0, start) + val + $el.value.slice(end)

      this.select(
        (isSelect ? start : start + val.length) + offset,
        start + val.length - offset
      )
      $el.scrollTop = scrollTop
      $el.focus()
    } else {
      $el.value += val
      $el.focus()
    }
    this.#updatePreview()
    this.$emit('input')
  }

  /**
   * [selection 获取选中的文本]
   * @param  {[type]} forceHoleLine [是否强制光标所在的整行文本]
   */
  selection(forceHoleLine = false) {
    let $el = this.$refs.editor
    let start = $el.selectionStart
    let end = $el.selectionEnd

    if (end) {
      //强制选择整行
      if (forceHoleLine) {
        start = $el.value.slice(0, start).lastIndexOf('\n')

        let tmpEnd = $el.value.slice(end).indexOf('\n')
        tmpEnd = tmpEnd < 0 ? $el.value.slice(end).length : tmpEnd

        start += 1 // 把\n加上
        end += tmpEnd

        $el.selectionStart = start
        $el.selectionEnd = end
      }
    } else {
      //强制选择整行
      if (forceHoleLine) {
        end = $el.value.indexOf('\n')
        end = end < 0 ? $el.value.length : end
        $el.selectionEnd = end
      }
    }
    $el.focus()
    return $el.value.slice(start, end)
  }

  select(start = 0, end = 0) {
    let $el = this.$refs.editor
    $el.selectionStart = start
    $el.selectionEnd = end
  }

  // 设置光标
  cursor(pos) {
    this.select(pos, pos)
  }

  #cursorMove(step) {
    let $el = this.$refs.editor
    let pos = (step < 0 ? $el.selectionStart : $el.selectionEnd) || 0
    pos += step

    if (step === 0) {
      return
    }

    if (pos < 0) {
      pos = 0
    }
    this.cursor(pos)
  }

  // 获取光标处的字符
  #getCursorText(n) {
    let $el = this.$refs.editor
    let start = $el.selectionStart
    let pos = $el.selectionEnd

    if (n < 0) {
      pos = start - 1
    }
    return this.value[pos] || ''
  }

  #handleKeydown(ev) {
    let $el = this.$refs.editor
    let wrapTxt = this.selection() || ''
    let selected = wrapTxt.length > 0
    let newTxt = ''

    if (this.readOnly || this.disabled) {
      return
    }

    switch (ev.keyCode) {
      //tab键改为插入2个空格,阻止默认事件,防止焦点失去
      case 9:
        ev.preventDefault()
        if (ev.shiftKey && !selected) {
          let pos = $el.selectionStart
          let line = this.selection(true)
          newTxt = line.replace(/^\s{2}/, '')
          // 防止无法往左取消缩进时, 选中整行
          if (line === newTxt) {
            this.cursor(pos)
            break
          }

          pos -= 2
          this.insert(newTxt, selected)
          this.cursor(pos)
        } else {
          newTxt = wrapTxt
            .split('\n')
            .map(function (it) {
              return ev.shiftKey ? it.replace(/^\s{2}/, '') : '  ' + it
            })
            .join('\n')

          if (newTxt === wrapTxt) {
            break
          }
          this.insert(newTxt, selected)
        }

        break

      // 退格键, 遇到成对的符号时,同时删除
      case 8:
        if (!selected) {
          let pos = $el.selectionStart
          let prev = this.#getCursorText(-1)
          let next = this.#getCursorText(1)

          if (
            (prev === next && (prev === '"' || prev === "'")) ||
            (prev === next && prev === '`') ||
            (prev === '[' && next === ']') ||
            (prev === '{' && next === '}') ||
            (prev === '(' && next === ')')
          ) {
            this.select(pos - 1, pos + 1)
          }
        }
        break

      // (
      case 57:
        if (ev.shiftKey) {
          ev.preventDefault()

          this.insert('(' + wrapTxt + ')', selected, (selected && 1) || 0)
          this.#cursorMove(selected ? 0 : -1)
        }
        break
      // )
      case 48:
        if (ev.shiftKey) {
          let prev = this.#getCursorText(-1)
          let next = this.#getCursorText(1)
          if (prev === '(' && next === ')') {
            ev.preventDefault()
            this.#cursorMove(1)
          }
        }
        break

      case 219: // [ & {
        ev.preventDefault()
        this.insert(
          ev.shiftKey ? `{${wrapTxt}}` : `[${wrapTxt}]`,
          selected,
          selected ^ 0
        )
        this.#cursorMove(selected ? 0 : -1)
        break
      // } & ]
      case 221: {
        let prev = this.#getCursorText(-1)
        let next = this.#getCursorText(1)
        if (ev.shiftKey) {
          if (prev === '{' && next === '}') {
            ev.preventDefault()
            this.#cursorMove(1)
          }
        } else {
          if (prev === '[' && next === ']') {
            ev.preventDefault()
            this.#cursorMove(1)
          }
        }
        break
      }

      //  `
      case 192:
      // ' & "
      case 222:
        if (ev.shiftKey && ev.keyCode === 192) {
          break
        } else {
          let val = ev.keyCode === 192 ? '`' : ev.shiftKey ? '"' : "'"
          let prev = this.#getCursorText(-1)
          let next = this.#getCursorText(1)

          if (prev === '' && next === val) {
            break
          }
          ev.preventDefault()

          if (selected) {
            this.insert(val + wrapTxt + val, true, 1)
          } else {
            if (prev === next && prev === val) {
              this.#cursorMove(1)
            } else {
              this.insert(val + wrapTxt + val)
              this.#cursorMove(-1)
            }
          }
          break
        }
    }
  }

  #syncScrollToPreview(ev) {
    if (this.previewEnabled) {
      let { editor, view } = this.$refs
      let st = editor.scrollTop
      let eh = editor.clientHeight
      let sh = editor.scrollHeight
      let veh = view.clientHeight
      let vsh = view.scrollHeight
      this.$refs.view.scrollTop = (st / (sh - eh)) * (vsh - veh)
    }
  }

  #fixedPadding() {
    if (this.clientHeight > 64) {
      let pb = ~~(this.clientHeight * 0.6) - (this.#toolbar.length ? 34 : 0)
      pb = pb < 64 ? 64 : pb
      this.$refs.editor.style.paddingBottom = pb + 'px'
    } else {
      this.$refs.editor.style.cssText = 'padding-bottom:;'
    }
  }

  mounted() {
    this.#fixedPadding()
    this._clickoutsideFn = outsideClick(this, _ => this.#hideLayers())

    this.__observer = new ResizeObserver(this.#fixedPadding.bind(this))
    this.__observer.observe(this)
  }

  unmounted() {
    this.__observer?.disconnect()
    clearOutsideClick(this._clickoutsideFn)
  }

  render() {
    let hasTable = this.#toolbar.includes('table')
    let hasHeader = this.#toolbar.includes('header')
    let hasLink = this.#toolbar.includes('link')

    return html`
      <div class="meditor">
        <header
          class=${classMap({ toolbar: true, active: this.#toolbar.length })}
          @click=${this.#toolbarClick}
        >
          ${this.#toolbar.map(it => {
            return html`<span data-act=${it}>
              <svg class="icon" viewBox="0 0 1024 1024">
                <path d=${ICONS[it]} />
              </svg>
              ${it === 'image' || it === 'attach'
                ? html`<input
                    type="file"
                    data-type=${it}
                    accept=${it === 'image' ? 'image/*' : '*/*'}
                    :disabled=${this.readOnly || this.disabled}
                    @change=${this.#handleFile}
                  />`
                : ''}
            </span>`
          })}
        </header>
        <div class="editor-outbox">
          <textarea
            ref="editor"
            class="editor"
            spellcheck="false"
            placeholder=${this.placeholder}
            autofocus=${this.autofocus}
            :readOnly=${this.readOnly}
            :disabled=${this.disabled}
            @click=${this.#hideLayers}
            @keydown=${this.#handleKeydown}
            @paste.prevent=${this.#handlePaste}
            @input=${this.#updatePreview}
            @scroll=${this.#syncScrollToPreview}
          ></textarea>

          <wc-markd ref="view" class="preview"></wc-markd>
        </div>

        ${hasHeader
          ? html`<div
              class="font-layer noselect"
              ref="header"
              @click=${this.#setHeaderLevel}
            >
              <span data-value="1">一级标题</span>
              <span data-value="2">二级标题</span>
              <span data-value="3">三级标题</span>
              <span data-value="4">四级标题</span>
              <span data-value="5">五级标题</span>
              <span data-value="6">六级标题</span>
            </div>`
          : ''}
        ${hasTable
          ? html`<div
              class="table-layer noselect"
              ref="table"
              @click=${this.#insertTable}
              @mousemove=${this.#tableSelect}
              @mouseleave=${this.#tableSelect}
            >
              ${range(81).map((_, n) => html`<span data-idx=${n}></span>`)}
            </div>`
          : ''}
        ${hasLink
          ? html`<div class="link-layer noselect" ref="link">
              <wc-input ref="linkinput" placeholder="请输入链接地址"></wc-input>
              <wc-button size="m" @click=${this.#insertLink}>插入</wc-button>
            </div>`
          : ''}
      </div>
    `
  }
}

MEditor.reg('meditor')
