/**
 * {选项卡组件}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/03/06 15:17:25
 */

import { css, html, raw, bind, Component, nextTick, classMap } from 'wkit'
import '../icon/index.js'
import { gzip } from '@bytedo/gzip'
import { colorHtml, colorJs, colorCss } from '../code/colorful.js'

const template = `<!doctype html>
<html lang="zh-CN">
  <head>
  <meta charset="utf-8">
  <link rel="stylesheet" href="//jscdn.ink/@bd/ui/latest/css/reset-basic.css">
  <style>body {padding:32px;}</style>
  <style name="css"></style>
  <script type="importmap">{"imports":{"es.shim":"//jscdn.ink/es.shim/latest/index.js","vue":"//jscdn.ink/vue/3.2.47/vue.esm-browser.prod.js","vue-router":"//jscdn.ink/@bytedo/vue-router/4.1.6/vue-router.js","fetch":"//jscdn.ink/@bytedo/fetch/latest/next.js","wkit":"//jscdn.ink/wkit/latest/index.js"}}</script>
  <script name="js" type="module"></script>
  </head>
  <body></body>
</html>`

function trim(str) {
  return str
    .trim()
    .replace(/&lt;/g, '<')
    .replace(/&gt;/g, '>')
    .replace(/&amp;/g, '&')
}

class Sandbox extends Component {
  static props = {
    tab: { type: String, default: 'preview', attribute: false },
    playground: {
      type: String,
      default: 'https://bd-js.github.io/playground.html',
      attribute: false
    }
  }

  static styles = css`
    :host {
      display: flex;
      flex-direction: column;
      border: 1px solid var(--color-plain-3);
      font-size: 14px;
      background: #fff;
    }

    .navs {
      position: relative;
      display: flex;
      width: 100%;
      height: 38px;
      color: var(--color-dark-1);
      background: var(--color-plain-1);
      box-shadow: inset 0 -1px 0 0px var(--color-plain-3);
      user-select: none;
    }

    .label {
      position: relative;
      display: inline-flex;
      align-items: center;
      justify-content: space-between;
      max-width: 120px;
      padding: 0 16px;
      border-right: 1px solid var(--color-plain-3);
      border-bottom: 1px solid var(--color-plain-3);
      background: var(--color-plain-1);
      --wc-icon-size: 16px;
      cursor: pointer;

      &:hover:not([disabled]),
      &.active {
        color: var(--color-teal-1);
      }
      &.active {
        background: #fff;
        border-bottom-color: transparent;
      }

      &[disabled] {
        opacity: 0.6;
        cursor: not-allowed;
      }
    }

    .open {
      position: absolute;
      right: 12px;
      top: 8px;
    }

    .content {
      flex: 1;
    }

    iframe {
      display: none;
      width: 100%;
      border: 0;
      background: #fff;
    }
  `

  #cache = {
    preview: { disabled: false, code: '' },
    javascript: { disabled: true, code: '' },
    html: { disabled: true, code: '' },
    css: { disabled: true, code: '' }
  }

  #created = false

  selectTab(ev) {
    let elem = ev.target
    let key

    if (elem === ev.currentTarget) {
      return
    }

    if (elem.tagName === 'LABEL') {
      key = elem.dataset.key

      if (key === this.tab) {
        return
      }

      if (this.#cache[key].disabled) {
        return
      }

      this.#cache[this.tab].panel.style.display = ''
      this.#cache[key].panel.style.display = 'block'

      this.tab = key
    } else {
      window.open(this.playground + '#' + gzip(this.code))
    }
  }

  get code() {
    let { javascript, html, css } = this.#cache
    let code = { js: javascript.code, html: html.code, css: css.code }
    return JSON.stringify(code)
  }

  created() {
    bind(this.root, 'slotchange', ev => {
      let slot = ev.target.assignedElements().pop()

      // 移除不合法的子组件

      if (slot.tagName !== 'WC-LANG') {
        return
      }

      let lang = slot.getAttribute('slot')

      if (lang) {
        this.#cache[lang].disabled = false
        this.#cache[lang].code = slot.code
        this.#cache[lang].panel = slot
      }

      this.updatePreview(lang, slot.code)
      this.$requestUpdate()
    })
  }

  updatePreview(lang, code) {
    let doc = this.$refs.preview.contentDocument
    if (this.#created) {
      switch (lang) {
        case 'css':
          doc.head.children.namedItem('css').textContent = code
          break
        case 'javascript':
          doc.head.children.namedItem('js').textContent = code
          break
        case 'html':
          doc.body.innerHTML = code
          break
      }
    }
  }

  contentLoaded(ev) {
    let elem = ev.target
    let body = elem.contentDocument.body
    elem.style.height = body.scrollHeight + 10 + 'px'
  }

  renderTemplate() {
    let doc = this.$refs.preview.contentDocument
    try {
      doc.open()
      doc.write(template)
      doc.close()
      this.#created = true
    } catch (e) {}
  }

  mounted() {
    //
    let target = this.$refs.preview
    this.#cache.preview.panel = target
    this.renderTemplate()
    target.style.display = 'block'

    this.__observer = new ResizeObserver(_ => this.contentLoaded({ target }))
    this.__observer.observe(target.contentDocument.body)
  }

  unmounted() {
    this.__observer?.disconnect()
  }

  render() {
    let labels = ['preview', 'javascript', 'html', 'css']

    return html`
      <header class="navs" ref="navs" @click=${this.selectTab}>
        ${labels.map(
          it => html`
            <label
              class=${classMap({
                label: true,
                active: it === this.tab
              })}
              data-key=${it}
              disabled=${this.#cache[it].disabled}
            >
              ${it}
            </label>
          `
        )}
        <wc-icon
          title="在playground中打开"
          class="open"
          name="fly"
          size="s"
        ></wc-icon>
      </header>
      <div class="content">
        <iframe ref="preview" @load=${this.contentLoaded}></iframe>
        <slot name="javascript"></slot>
        <slot name="html"></slot>
        <slot name="css"></slot>
      </div>
    `
  }
}

class Lang extends Component {
  static props = {
    code: {
      type: String,
      default: '',
      attribute: false,
      observer(v) {
        this.setCode(v.trim())
      }
    }
  }

  static styles = [
    css`
      :host {
        display: none;
        width: 100%;
        background: #fff;
        color: var(--color-dark-1);
      }
    `,
    css`
      .code-block {
        display: flex;
        flex-direction: column;
        overflow: hidden;
        overflow-y: auto;
        padding: 6px 0;
        line-height: 20px;
        font-size: 14px;
        color: var(--color-dark-1);
        cursor: text;
        counter-reset: code;

        > code {
          display: block;
          position: relative;
          min-height: 20px;
          padding: 0 8px 0 45px;
          font-family: Menlo, Monaco, Consolas, 'Courier New', monospace;
          white-space: pre-wrap;
          word-break: break-word;

          &::before {
            position: absolute;
            left: 0;
            width: 40px;
            height: 100%;
            padding-right: 5px;
            text-align: right;
            color: var(--color-grey-1);
            content: counter(code);
            counter-increment: code;
          }

          .t-tag,
          .t-keyword,
          .t-const {
            color: var(--color-red-2);
          }
          .t-attr,
          .t-fn,
          .t-var {
            color: var(--color-blue-2);
          }
          .t-str {
            color: var(--color-green-2);
          }
          .t-comment {
            color: var(--color-grey-2);
          }

          .t-type {
            font-weight: bold;
            color: var(--color-orange-1);
          }
          .t-num,
          .t-buildin {
            color: #a46ad3;
          }
          .t-link {
            text-decoration: underline;
            color: var(--color-grey-2);
          }
          .t-buildin,
          .t-keyword,
          .t-type,
          .t-link,
          .t-comment {
            font-style: italic;
          }
        }
      }
    `
  ]

  #code = []

  setCode(txt, a) {
    let lang = this.getAttribute('slot')

    switch (lang) {
      case 'javascript':
        txt = colorJs(txt)
        break

      case 'html':
        txt = colorHtml(txt)
        break

      case 'css':
        txt = colorCss(txt)
        break
    }

    this.#code = txt.split('\n')
  }

  mounted() {
    var txt = this.innerHTML || this.textContent
    txt = txt.trim().replace(/^[\r\n]|\s{2,}$/g, '')

    if (txt.startsWith('<xmp>') && txt.endsWith('</xmp>')) {
      txt = txt.slice(5, -6).trim()
    } else if (this.firstElementChild?.tagName === 'TEXTAREA') {
      txt = this.firstElementChild.value.trim()
    } else if (txt.startsWith('<pre>') && txt.endsWith('</pre>')) {
      txt = trim(txt.slice(5, -6))
    } else {
      txt = trim(txt)
    }

    this.textContent = ''
    if (txt) {
      nextTick(_ => {
        this.code = txt
      })
    }
  }

  render() {
    return html`
      <div class="code-block">
        ${this.#code.map(s => html`<code>${raw(s)}</code>`)}
      </div>
    `
  }
}

Sandbox.reg('sandbox')
Lang.reg('lang')
