/**
 * 轮播图组件
 * @author chensbox<chensbox@foxmail.com>
 * @date 2023/03/26 16:14:10
 */

import { css, html, Component, nextTick, styleMap } from 'wkit'
import '../icon/index.js'
const CARD_SCALE = 0.83

function createWatcher(object, key, effect) {
  let value
  Object.defineProperty(object, key, {
    get() {
      return value
    },
    set(val) {
      value = val
      effect.call(object, val)
    }
  })
}

class Swipe extends Component {
  static props = {
    initialIndex: {
      type: Number,
      default: 0
    },
    height: {
      type: String,
      default: 200,
      observer(val) {
        this.style.height = `${val}px`
      }
    },
    trigger: {
      type: String,
      default: 'hover' // hover or click
    },
    autoplay: {
      type: Boolean,
      default: true,
      observer(val) {
        val ? this.startTimer() : this.pauseTimer()
      }
    },
    interval: {
      type: Number,
      default: 3000
    },
    indicator: {
      type: Boolean,
      default: true
    },
    arrow: {
      type: String,
      default: 'hover' // hover or alway or none
    },
    type: {
      type: String,
      default: '', // '' or card
      attributes: false
    },
    loop: {
      type: Boolean,
      default: true
    },
    direction: {
      type: String,
      default: 'horizontal' // horizontal or  vertical
    },
    pauseOnHover: {
      type: Boolean,
      default: true
    },
    space: {
      type: Number,
      default: 0
    }
  }

  static styles = css`
    :host {
      overflow: hidden;
      position: relative;
      display: block;
      width: 100%;
    }
    :host([direction='vertical']) {
      .indicator {
        flex-direction: column;
        left: calc(100% - 26px);
        top: 50%;
        height: 80%;
        width: 26px;
        transform: translatey(-50%);
        li {
          width: 26px;
          height: 24px;
          margin: 5px 0;
          &::after {
            position: absolute;
            content: '';
            left: calc(50% - 2px);
            top: 0;
            height: 100%;
            width: 2px;
            background-color: #fff;
            opacity: 0.48;
          }
          &.active::after {
            opacity: 1;
          }
        }
      }
    }
    .toggle-btn {
      z-index: 4;
      user-select: none;
      cursor: pointer;
      position: absolute;
      display: flex;
      align-items: center;
      justify-content: center;
      opacity: 0;
      width: 33px;
      height: 33px;
      font-size: 24px;
      color: #fff;
      background-color: #606266;
      border-color: #fff;
      border-radius: 50%;
      --wc-icon-size: 12px;
      transition: opacity ease-in-out 0.2s;
      &:hover {
        opacity: 1 !important;
      }
    }
    .left,
    .right {
      top: 50%;
      transform: translateY(-50%);
    }
    .left {
      left: 3%;
    }
    .right {
      right: 3%;
    }
    .indicator {
      z-index: 10;
      display: flex;
      align-items: center;
      justify-content: center;
      position: absolute;
      width: 80%;
      bottom: 0px;
      left: 50%;
      transform: translateX(-50%);
      li {
        position: relative;
        width: 24px;
        height: 26px;
        margin: 0 5px;
        list-style: none;
        &::after {
          position: absolute;
          content: '';
          left: 0;
          top: calc(50% - 2px);
          width: 100%;
          height: 2px;
          background-color: #fff;
          opacity: 0.48;
        }
        &:hover {
          cursor: pointer;
        }
        &.active::after {
          opacity: 1;
        }
      }
    }
  `

  items = []
  activeIndex = 0
  stamp = 0

  updateItems() {
    this.items = Array.from(this.children).filter(
      item => item.tagName === 'WC-SWIPE-ITEM'
    )
    nextTick(() => this.$requestUpdate())
  }
  playSlides() {
    let oldIndex = this.activeIndex
    if (this.activeIndex < this.items.length - 1) {
      this.activeIndex++
    } else if (this.loop) {
      this.activeIndex = 0
    }
    this.resetItemPosition(oldIndex)
  }
  startTimer() {
    if (!this.interval || this.interval <= 0 || !this.autoplay || this.timer) {
      return
    }
    this.timer = setInterval(this.playSlides.bind(this), this.interval)
  }
  pauseTimer() {
    if (this.timer) {
      clearInterval(this.timer)
      this.timer = null
    }
  }
  resetTimer() {
    this.pauseTimer()
    this.startTimer()
  }
  resetItemPosition(oldIndex) {
    this.items.forEach((item, index) => {
      item.translateItem(index, this.activeIndex, oldIndex)
    })
    nextTick(() => this.$requestUpdate())
  }
  setActiveItem(index) {
    const now = Date.now()
    if (now - this.stamp < 300) {
      return
    }
    this.stamp = now

    if (typeof index === 'string') {
      const filteredItems = this.items.filter(item => item.name === index)
      if (filteredItems.length > 0) {
        index = this.items.indexOf(filteredItems[0])
      }
    }

    index = Number(index)
    if (isNaN(index) || index !== Math.floor(index)) {
      return
    }

    const length = this.items.length
    const oldIndex = this.activeIndex
    if (index < 0) {
      this.activeIndex = this.loop ? length - 1 : 0
    } else if (index >= length) {
      this.activeIndex = this.loop ? 0 : length - 1
    } else {
      this.activeIndex = index
    }
    if (oldIndex !== this.activeIndex) {
      this.resetItemPosition(oldIndex)
    }

    this.resetTimer()
    nextTick(() => this.$requestUpdate())
  }
  handleIndicatorHover(event) {
    let index = +event.target.getAttribute('index')
    if (this.trigger === 'hover' && index !== this.activeIndex) {
      this.setActiveItem(index)
    }
  }
  handleClickIndicator(event) {
    let index = +event.target.getAttribute('index')
    if (this.trigger === 'click' && index !== this.activeIndex) {
      this.setActiveItem(index)
    }
  }
  prev() {
    this.setActiveItem(this.activeIndex - 1)
  }
  next() {
    this.setActiveItem(this.activeIndex + 1)
  }
  showArrow() {
    this.showArrowBtn = true
    this.$requestUpdate()
  }
  hideArrow() {
    this.showArrowBtn = false
    this.$requestUpdate()
  }
  handleSlotChange() {
    this.updateItems()
    this.resetItemPosition()
  }

  mounted() {
    this.updateItems()
    nextTick(() => {
      let initialIndex = this.initialIndex
      if (initialIndex >= 0 && initialIndex < this.items.length) {
        this.activeIndex = initialIndex
      }
      this.resizeObserve = new ResizeObserver(this.resetItemPosition.bind(this))
      this.resizeObserve.observe(this)
      this.startTimer()
    })

    if (this.pauseOnHover) {
      this.$on('mouseenter', () => this.pauseTimer())
      this.$on('mouseleave', () => this.startTimer())
    }

    if (this.arrow === 'hover') {
      this.$on('mouseenter', () => this.showArrow())
      this.$on('mouseleave', () => this.hideArrow())
    } else if (this.arrow === 'alway') {
      this.showArrow()
    }
  }

  unmounted() {
    this.resizeObserve.disconnect()
    this.resizeObserve = null
    this.pauseTimer()
  }

  render() {
    let styles = styleMap({
      opacity: this.showArrowBtn ? 0.6 : 0
    })

    let toggleBtns =
      this.arrow !== 'none' && this.direction === 'horizontal'
        ? html`
            <div class="left toggle-btn" style=${styles} @click=${this.prev}>
              <wc-icon name="left"></wc-icon>
            </div>
            <div class="right toggle-btn" style=${styles} @click=${this.next}>
              <wc-icon name="right"></wc-icon>
            </div>
          `
        : ''

    let indicator = this.indicator
      ? html`
          <ul class="indicator" @click=${this.handleClickIndicator}>
            ${this.items.map(
              (item, index) =>
                html` <li
                  index=${index}
                  class=${this.activeIndex === index ? 'active' : ''}
                  @mouseenter=${this.handleIndicatorHover}
                ></li>`
            )}
          </ul>
        `
      : ''

    return html`
      <div class="container" style=${`height:${this.height}px`}>
        <slot @slotchange=${this.handleSlotChange}></slot>
        ${toggleBtns}
      </div>
      ${indicator}
    `
  }
}

class SwipeItem extends Component {
  static props = {
    name: { String, attributes: false }
  }

  static styles = css`
    :host {
      position: absolute;
      left: 0;
      top: 0;
      display: none;
      width: 100%;
      height: 100%;
      transition: transform ease-in-out 0s;
    }
    :host([ready]) {
      display: block;
    }
    :host([animating]) {
      transition-duration: 400ms;
    }
    :host([card]) {
      width: 50%;
      transition-duration: 400ms;
      transition-property: transform, opacity, filter;
      cursor: pointer;
    }
  `

  static watch = {
    translate(val) {
      const translateType =
        this.$parent.direction === 'horizontal' ? 'translateX' : 'translateY'
      this.style.transform = `${translateType}(${val}px) scale(${this.scale})`
    },

    ready(val) {
      val ? this.setAttribute('ready', '') : this.removeAttribute('ready')
    },

    animating(val) {
      val
        ? this.setAttribute('animating', '')
        : this.removeAttribute('animating')
    },

    active(val) {
      if (this.$parent.type !== 'card') {
        return
      }

      const style = {
        opacity: 1,
        'z-index': 1,
        filter: 'brightness(0.8)'
      }

      if (this.inStage) {
        style['z-index'] = 1
      } else {
        style['z-index'] = -1
        style.opacity = 0
      }

      if (val) {
        style['z-index'] = 2
        style.filter = 'brightness(1)'
      }

      Object.assign(this.style, style)
    }
  }

  processIndex(index, activeIndex, length) {
    if (activeIndex === 0 && index === length - 1) {
      return -1
    } else if (activeIndex === length - 1 && index === 0) {
      return length
    } else if (index < activeIndex - 1 && activeIndex - index >= length / 2) {
      return length + 1
    } else if (index > activeIndex + 1 && index - activeIndex >= length / 2) {
      return -2
    }
    return index
  }

  calcCardTranslate(index, activeIndex) {
    let space = 0
    if (this.$parent.space) {
      space = 0.66 + +this.$parent.space / 100
    }
    const parentWidth = this.$parent.offsetWidth
    if (this.inStage) {
      return (
        (parentWidth * ((2 + space - CARD_SCALE) * (index - activeIndex) + 1)) /
        4
      )
    } else if (index < activeIndex) {
      return (-(1 + CARD_SCALE) * parentWidth) / 4
    } else {
      return ((3 + CARD_SCALE) * parentWidth) / 4
    }
  }

  calcTranslate(index, activeIndex, isVertical) {
    const distance = this.$parent[isVertical ? 'offsetHeight' : 'offsetWidth']
    return distance * (index - activeIndex)
  }

  translateItem(index, activeIndex, oldIndex) {
    const parentType = this.$parent.type
    const parentDirection = this.$parent.direction
    const length = this.$parent.items.length

    if (parentType !== 'card' && oldIndex !== undefined) {
      this.animating = index === activeIndex || index === oldIndex
    }

    if (index !== activeIndex && length > 2 && this.$parent.loop) {
      index = this.processIndex(index, activeIndex, length)
    }

    if (parentType === 'card') {
      if (parentDirection === 'vertical') {
        console.warn(
          '[Warn][Carousel]vertical direction is not supported in card mode'
        )
      }

      this.inStage = Math.round(Math.abs(index - activeIndex)) <= 1
      this.active = index === activeIndex
      this.scale = this.active ? 1 : CARD_SCALE
      this.animating = true
      this.translate = this.calcCardTranslate(index, activeIndex)
    } else {
      this.active = index === activeIndex
      const isVertical = parentDirection === 'vertical'
      this.scale = 1
      this.translate = this.calcTranslate(index, activeIndex, isVertical)
    }

    !this.ready && (this.ready = true)
  }

  created() {
    const { watch } = this.constructor
    for (const key in watch) {
      createWatcher(this, key, watch[key])
    }
  }

  mounted() {
    this.$parent = this.parentNode
    if (this.$parent.type === 'card') {
      this.setAttribute('card', '')
      const index = this.$parent.items.indexOf(this)
      this.$on('click', () => this.$parent.setActiveItem(index))
    }
  }
}

Swipe.reg('swipe')
SwipeItem.reg('swipe-item')
